#ifndef __MATRIX4D_H__
#define __MATRIX4D_H__

#include "mathlibx.h"
#include "vector.h"
class Quater;

// Column Order Elements
// Based on OpenGL Description
//
//            4 x 4        =                1 x 16
//
// M = | m0  m4  m8  m12 | = { m0, m1, m2,  m3,  m4,  m5,  m6,  m7, 
//     | m1  m5  m9  m13 |     m8, m9, m10, m11, m12, m13, m14, m15 }
//     | m2  m6  m10 m14 |
//     | m3  m7  m11 m15 |

//////////////////////////////////////////////////
// Operations Supported
//
// Matrix mA, mB, mC;// Matrices
// Vector u, v, w;   // Vectors
// FLOAT f;          // Scalars
// BOOL b;           // Booleans
//
// f = v[0];       // LHS Index Operation
// v[0] = f;       // LHS Index Operation
// f = v(0);       // RHS Index Operation
// v(0) = f;       // ILLEGAL -- RHS (right hand side) only!!!
//
// u = v;
// u += v;
// u -= v;
// u *= f;         // Scalar Multiplication
// u = -v;
// u = v + w;
// u = v - w;
// u = v * w;      // Cross Product
// u = v.Cross(w); // Cross Product
// f = v.Dot(w);   // Dot Product
// u = v * f;      // Scalar Multiplication
// u = f * v;      // Scalar Multiplication
// u = v / f;      // Scalar Division
// u = f / v;      // Scalar Inversion
//

class Matrix
{
public:
  Matrix();
  ~Matrix();

public:
  inline UINT_32 Index(UINT_32 uiRow, UINT_32 uiCol) const;
  inline VOID Set(const MATRIX4D m);
  inline VOID Get(MATRIX4D m) const;
  inline const FLOAT *Get() const;
  inline VOID Set(UINT_32 uiRow, UINT_32 uiCol, FLOAT f);
  inline FLOAT Get(UINT_32 uiRow, UINT_32 uiCol) const;

  // Initialization Methods
public:
  VOID SetNull();
  VOID SetIdentity();

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 5: You need to implement these!!!
// 
  VOID SetTranslate(const Vector &v);
  // This is useful with the Rodrigues Formula for general axis rotations
  VOID SetCross(const Vector &v); 
  // Ideally use the Rodrigues Formula to form the general axis rotation
  VOID SetRotate(const Vector &vW, DOUBLE dTheta);
  // Change of Basis rotation, which is useful to creat the LookAt transform
  // mGa_b these are b's basis vectors represented in a's coordinate system
  VOID SetRotate(const Vector &vEx, const Vector &vEy, const Vector &vEz);
// 
// END  : ASSIGNMENT 5: You need to implement these!!!
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 5: Not necessary, but if you really like Quaternions ...
// 
  VOID SetRotate(const Quater &q);
// 
// END  : ASSIGNMENT 5: Not necessary, but if you really like Quaternions ...
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 5: You need to implement these!!!
// 
  VOID SetRotateX(DOUBLE dTheta);
  VOID SetRotateY(DOUBLE dTheta);
  VOID SetRotateZ(DOUBLE dTheta);
  VOID SetRotateZ(FLOAT sint, FLOAT cost);
  VOID SetScale(const Vector &v);
  VOID SetScale(const FLOAT f);
  VOID SetShearXY(const Vector &v);
  VOID SetShearYZ(const Vector &v);
  VOID SetShearXZ(const Vector &v);
// 
// END  : ASSIGNMENT 5: You need to implement these!!!
//////////////////////////////////////////////////////////////////////////////

  // Transform from a World Window (WW) to a screne ViewPort (VP)
  VOID SetViewPort(const Vector &vVPMin, const Vector &vVPMax, 
		   const Vector &vWWMin, const Vector &vWWMax,
		   FLOAT fDYDX = 1.0);

  // Decomposition Methods
public:
  FLOAT Trace() const;
  FLOAT Determinant() const;

  // Operations on this Matrix
public:
  VOID Transpose();

  // Operations which create new Matrices
public:  
  Matrix &operator=(const Matrix& m);
  Matrix &operator+=(const Matrix& m);
  Matrix &operator-=(const Matrix& m);
  Matrix &operator*=(const FLOAT f);
  Matrix &operator/=(const FLOAT f);

  Matrix operator*(const Matrix& m) const;
  Vector operator*(const Vector& v) const;
  Point operator*(const Point& pt) const;

public:
  VOID Display() const;

protected:
  MATRIX4D m_m; // Column Order Elements
};

//////////////////////////////////////////////////////////////////////
// External Operators
//

inline Matrix operator+(const Matrix& mA, const Matrix& mB);
inline Matrix operator-(const Matrix& mA, const Matrix& mB);
inline Matrix operator*(const Matrix& m, const FLOAT f);
inline Matrix operator*(const FLOAT f, const Matrix& m);
inline Matrix operator/(const Matrix& m, const FLOAT f);

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 5: Hey look at this it will speed up your normal xforming
// 

// Row Operations -- useful for transforming normals
// w == v * M == Transpose(M) * v 
Vector operator*(const Vector& v, const Matrix &m);

// 
// END  : ASSIGNMENT 5: Hey look at this it will speed up your normal xforming
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline UINT_32 Matrix::Index(UINT_32 uiRow, UINT_32 uiCol) const
{
  return 4*uiCol + uiRow;
}

inline VOID Matrix::Set(const MATRIX4D m)
{
  memcpy(m_m, m, sizeof(MATRIX4D));
}

inline VOID Matrix::Get(MATRIX4D m) const
{
  memcpy(m, m_m, sizeof(MATRIX4D));
}

inline const FLOAT *Matrix::Get() const
{
  return m_m;
}

inline VOID Matrix::Set(UINT_32 uiRow, UINT_32 uiCol, FLOAT f)
{
  ASSERT( uiRow < 4 && uiCol < 4 );
  m_m[Index(uiRow, uiCol)] = f;
}

inline FLOAT Matrix::Get(UINT_32 uiRow, UINT_32 uiCol) const
{
  ASSERT( uiRow < 4 && uiCol < 4 );
  return m_m[Index(uiRow, uiCol)];
}

//////////////////////////////////////////////////////////////////////
// Implementation of External Functions
//

inline Matrix operator+(const Matrix& mA, const Matrix& mB)
{
  Matrix mResult;

  mResult = mA;
  mResult += mB;

  return mResult;
}

inline Matrix operator-(const Matrix& mA, const Matrix& mB)
{
  Matrix mResult;

  mResult = mA;
  mResult -= mB;

  return mResult;
}

inline Matrix operator*(const Matrix& m, const FLOAT f)
{
  Matrix mResult;

  mResult = m;
  mResult *= f;

  return mResult;
}

inline Matrix operator*(const FLOAT f, const Matrix& m)
{
  Matrix mResult;

  mResult = m;
  mResult *= f;

  return mResult;
}

inline Matrix operator/(const Matrix& m, const FLOAT f)
{
  Matrix mResult;

  mResult = m;
  mResult /= f;

  return mResult;
}

#endif // __MATRIX4D_H__










