#include "vector.h"
#include "point.h"
#include "quater.h"

Quater::Quater()
{
   SetIdentity();
}

Quater::Quater(FLOAT x, FLOAT y, FLOAT z, FLOAT w/* = 0.0f*/)
{
   m_q[0] = x;
   m_q[1] = y;
   m_q[2] = z;
   m_q[3] = w;
}

Quater::Quater(const FLOAT *v)
{
   ASSERT( v != NULL );

   m_q[0] = v[0];
   m_q[1] = v[1];
   m_q[2] = v[2];
   m_q[3] = v[3];
}

Quater::~Quater()
{
}

FLOAT Quater::Magnitude() const
{
  return (FLOAT)sqrt(m_q[0]*m_q[0] + m_q[1]*m_q[1] + m_q[2]*m_q[2] + m_q[3]*m_q[3]);
}

FLOAT Quater::Magn2() const
{
  return (FLOAT)(m_q[0]*m_q[0] + m_q[1]*m_q[1] + m_q[2]*m_q[2] + m_q[3]*m_q[3]);
}

VOID Quater::Null()
{
  memset(m_q, 0, sizeof(VECTOR4D));
}

VOID Quater::SetIdentity()
{
  memset(m_q, 0, sizeof(VECTOR4D));
  m_q[3] = 1.0;
}

Quater Quater::Inv() const
{
   FLOAT fMag;

   fMag = Magn2();

   if ( fMag <= 0.0f )
   {
      return *this;
   }

   return Conj()/fMag;
}

VOID Quater::GetRotation(Vector &vAxis, FLOAT &fAngle) const
{
  vAxis.Set(m_q[0], m_q[1], m_q[2]);

  if ( !vAxis.Normalize() )
    {
      vAxis.Set(0.0, 0.0, 1.0);
      fAngle = 0.0;
    }
  else
    {
      fAngle = (FLOAT)(2.0*acos(m_q[3]));
    }
}

BOOL Quater::SetRotation(const Vector &axis, FLOAT angle)
{
   FLOAT ca, sa;
   Vector naxis;

   naxis = axis;

   if (naxis.Normalize())
   {
      ca = (FLOAT)cos(angle/2);
      sa = (FLOAT)sin(angle/2);
      m_q[0] = naxis(0) * sa;
      m_q[1] = naxis(1) * sa;
      m_q[2] = naxis(2) * sa;
      m_q[3] = ca;
      return TRUE;
   } else {
      SetUnit();
      return FALSE;
   }
}

VOID Quater::SetRotation(const Vector &v1, const Vector &v2)
{
  Vector vA, vB, v;
  FLOAT cosu, cosu_2, sinu_2;

  vA = v1;
  vB = v2;
  
  vA.Normalize();
  vB.Normalize();

  cosu = (FLOAT)MIN(1.0, MAX(-1.0, vA.Dot(vB)));

  if ( cosu == -1.0 )
    {
      // Anti-parallel vectors
      UINT_32 uiMin;
      FLOAT fMin;
      
      sinu_2 = 1.0;
      cosu_2 = 0.0;

      fMin = vA(0);
      uiMin = 0;
      if ( vA(1) < fMin )
	{
	  fMin = vA(1);
	  uiMin = 1;
	}
      if ( vA(2) < fMin )
	{
	  fMin = vA(2);
	  uiMin = 2;
	}

      v[uiMin] = 1.0;
      v = v * vA;
    }
  else
    {
      v = vA * vB;

      cosu_2 = (FLOAT)(sqrt((1.0 + cosu) / 2.0));
      sinu_2 = (FLOAT)(sqrt((1.0 - cosu) / 2.0));
    }

  v.Normalize();
  m_q[0] = v[0]*sinu_2;
  m_q[1] = v[1]*sinu_2;
  m_q[2] = v[2]*sinu_2;
  m_q[3] = cosu_2;
}

Point Quater::RotatePoint(const Point &pt) const
{
   Quater qp, qr;

   qp.Set(pt);
   qr = ((*this) * qp * Inv());
   return Point(qr.m_q[0],qr.m_q[1],qr.m_q[2],qr.m_q[3]);
}

Vector Quater::RotateVector(const Vector &v) const
{
   Quater qv, qr;

   qv.Set(v);
   qr = ((*this) * qv * Inv());
   return Vector(qr.m_q[0],qr.m_q[1],qr.m_q[2],qr.m_q[3]);
}

Quater Quater::RotateQuat(const Quater &q) const
{
   return ((*this) * q * Inv());
}

BOOL Quater::Normalize()
{
   FLOAT fMag;

   fMag = Magnitude();

   if ( fMag <= 0.0f )
   {
      return FALSE;
   }

   m_q[0] = m_q[0] / fMag;
   m_q[1] = m_q[1] / fMag;
   m_q[2] = m_q[2] / fMag;
   m_q[3] = m_q[3] / fMag;

   return TRUE;
}

Quater &Quater::operator=(const Quater &v)
{
   m_q[0] = v(0);
   m_q[1] = v(1);
   m_q[2] = v(2);
   m_q[3] = v(3);

   return *this;
}

Quater &Quater::operator=(const Vector &v)
{
   m_q[0] = v(0);
   m_q[1] = v(1);
   m_q[2] = v(2);
   m_q[3] = v(3);

   return *this;
}

Quater &Quater::operator=(const Point &pt)
{
   m_q[0] = pt(0);
   m_q[1] = pt(1);
   m_q[2] = pt(2);
   m_q[3] = pt(3);

   return *this;
}

Quater &Quater::operator+=(const Quater &v)
{
   m_q[0] = m_q[0] + v(0);
   m_q[1] = m_q[1] + v(1);
   m_q[2] = m_q[2] + v(2);
   m_q[3] = m_q[3] + v(3);

   return *this;
}

Quater &Quater::operator-=(const Quater &v)
{
   m_q[0] = m_q[0] - v(0);
   m_q[1] = m_q[1] - v(1);
   m_q[2] = m_q[2] - v(2);
   m_q[3] = m_q[3] - v(3);

   return *this;
}

Quater &Quater::operator*=(const Quater &v)
{
   Quater q = (*this)*v;
   (*this) = q;
   return *this;
}

Quater &Quater::operator/=(const Quater &v)
{
   Quater q = (*this)*v;
   (*this) = q;
   return *this;
}

Quater &Quater::operator*=(const FLOAT f)
{
   m_q[0] = m_q[0] * f;
   m_q[1] = m_q[1] * f;
   m_q[2] = m_q[2] * f;
   m_q[3] = m_q[3] * f;

   return *this;
}

Quater &Quater::operator/=(const FLOAT f)
{
   m_q[0] = m_q[0] / f;
   m_q[1] = m_q[1] / f;
   m_q[2] = m_q[2] / f;
   m_q[3] = m_q[3] / f;

   return *this;
}

BOOL Quater::operator==(const Quater &v) const
{
   Quater diff;

   diff = (*this)-v;
   return ( diff.Magn2() <= EPSILON*EPSILON );
}

BOOL Quater::operator!=(const Quater &v) const
{
   return !(*this == v);
}

VOID Quater::Display() const
{
   UINT_32 i;

   printf("{ ");
   for ( i = 0; i < 4; i++)
   {
      printf("%6.3f ", (*this)(i));
   }
   printf("}\n");
}
