#ifndef __QUATER_H__
#define __QUATER_H__

#include "mathlibx.h"

//////////////////////////////////////////////////
// Operations Supported
//
// Quater u, v, w;   	// Quaternions
// Vector x, y;		// Vectors
// Point p;          	// Points
// FLOAT f;          	// Scalars
// BOOL b;           	// Booleans
//
// f = v[0];       // LHS Index Operation
// v[0] = f;       // LHS Index Operation
// f = v(0);       // RHS Index Operation
// v(0) = f;       // ILLEGAL -- RHS (right hand side) only!!!
//
// u = v
// u += v;
// u -= v;
// u *= v;
// u /= v;
// u *= f;         // Scalar Multiplication
// u /= f;         // Scalar Division

// u = v.Conj();   // Conjugate
// u = v.Inv();    // Multiplicative Inverse (u = 1/v)

// u = -v;
// u = v + w;
// u = v - w;
// u = v * w;      // Quaternion Product
// u = v / w;      // Quaternion Product
// u = v * f;      // Scalar Multiplication
// u = f * v;      // Scalar Multiplication
// u = v / f;      // Scalar Division
// u = f / v;      // Scalar Inversion
//
//
// Logic
// b = (u == v); // Quasi Equal allows for floating point inaccuracy
// b = (u != v); // Quasi Not Equal allows for floating point inaccuracy

//////////////////////////////////////////////////////////////////////
// Quater Class
//

class Point;
class Vector;

class Quater
{
public:
   Quater();
   Quater(FLOAT x, FLOAT y, FLOAT z, FLOAT w = 0.0f);
   Quater(const FLOAT *v);
   virtual ~Quater();

public:
   inline VOID Set(FLOAT x, FLOAT y, FLOAT z, FLOAT w = 0.0f);
   inline VOID Set(FLOAT f);
   inline VOID Set(const VECTOR4D v);
   inline VOID Set(const Vector v);
   inline VOID Set(const Point p);
   inline VOID Get(VECTOR4D v) const;
   inline const FLOAT *Get() const;

   inline VOID SetUnit();
   inline VOID SetZero();
   VOID GetRotation(Vector &vAxis, FLOAT &fAngle) const;
   BOOL SetRotation(const Vector &axis, FLOAT angle);
   VOID SetRotation(const Vector &v1, const Vector &v2);
   Point RotatePoint(const Point &pt) const;
   Vector RotateVector(const Vector &v) const;
   Quater RotateQuat(const Quater &q) const;
   inline Quater Conj() const;

public:
   FLOAT Magnitude() const;
   FLOAT Magn2() const;
   VOID Null();
   VOID SetIdentity();
   BOOL Normalize();

public:
   inline FLOAT &operator[](const UINT_32 uiIndex);
   inline FLOAT operator()(const UINT_32 uiIndex) const;

   Quater &operator=(const Quater &v);
   Quater &operator=(const Vector &v);
   Quater &operator=(const Point &pt);
   Quater &operator+=(const Quater &v);
   Quater &operator-=(const Quater &v);
   Quater &operator*=(const Quater &v);
   Quater &operator/=(const Quater &v);
   Quater &operator*=(const FLOAT f);
   Quater &operator/=(const FLOAT f);

   Quater Inv() const;

   BOOL operator==(const Quater &v) const;
   BOOL operator!=(const Quater &v) const;

public:
   VOID Display() const;

protected:
   VECTOR4D m_q;
};

//////////////////////////////////////////////////////////////////////
// External Operators
//

inline Quater operator-(const Quater &v);
inline Quater operator+(const Quater &v, const Quater &w);
inline Quater operator-(const Quater &v, const Quater &w);
inline Quater operator*(const Quater &v, const Quater &w);
inline Quater operator/(const Quater &v, const Quater &w);
inline Quater operator*(const Quater &v, const FLOAT f);
inline Quater operator*(const FLOAT f, const Quater &v);
inline Quater operator/(const Quater &v, const FLOAT f);
inline Quater operator/(const FLOAT f, const Quater &v);

//////////////////////////////////////////////////////////////////////////////
// Inlined Functions
//

inline VOID Quater::Set(FLOAT x, FLOAT y, FLOAT z, FLOAT w/* = 0.0f*/)
{
   m_q[0] = x;
   m_q[1] = y;
   m_q[2] = z;
   m_q[3] = w;
}

inline VOID Quater::Set(const VECTOR4D v)
{
   ASSERT( v != NULL );

   memcpy(m_q, v, sizeof(VECTOR4D));
}

inline VOID Quater::Set(FLOAT f)
{
   SetZero();
   m_q[3] = f;
}

inline VOID Quater::Set(const Vector v)
{
   m_q[0] = v(0);
   m_q[1] = v(1);
   m_q[2] = v(2);
   m_q[3] = v(3);
}

inline VOID Quater::Set(const Point p)
{
   m_q[0] = p(0);
   m_q[1] = p(1);
   m_q[2] = p(2);
   m_q[3] = p(3);
}

inline VOID Quater::Get(VECTOR4D v) const
{
   ASSERT( v != NULL );

   memcpy(v, m_q, sizeof(VECTOR4D));
}

inline const FLOAT *Quater::Get() const
{
   return m_q;
}

inline VOID Quater::SetUnit()
{
   m_q[0] = m_q[1] = m_q[2] = 0.0;
   m_q[3] = 1.0;
}

inline VOID Quater::SetZero()
{
   m_q[0] = m_q[1] = m_q[2] = m_q[3] = 0.0;
}

inline Quater Quater::Conj() const
{
   return Quater(-m_q[0], -m_q[1], -m_q[2], m_q[3]);
}

inline FLOAT &Quater::operator[](const UINT_32 uiIndex)
{
   return m_q[uiIndex];
}

inline FLOAT Quater::operator()(const UINT_32 uiIndex) const
{
   return m_q[uiIndex];
}

//////////////////////////////////////////////////////////////////////
// External Operator Implementations
//

inline Quater operator-(const Quater &v)
{
   return Quater(-v(0), -v(1), -v(2), -v(3));
}

inline Quater operator+(const Quater &v, const Quater &w)
{
   return Quater(v(0) + w(0), v(1) + w(1), v(2) + w(2), v(3) + w(3));
}

inline Quater operator-(const Quater &v, const Quater &w)
{
   return Quater(v(0) - w(0), v(1) - w(1), v(2) - w(2), v(3) - w(3));
}

inline Quater operator*(const Quater &v, const Quater &w)
{
   return Quater( v(0)*w(3) + v(1)*w(2) - v(2)*w(1) + v(3)*w(0),
		  -v(0)*w(2) + v(1)*w(3) + v(2)*w(0) + v(3)*w(1),		
		  v(0)*w(1) - v(1)*w(0) + v(2)*w(3) + v(3)*w(2),		
		  -v(0)*w(0) - v(1)*w(1) - v(2)*w(2) + v(3)*w(3));
}

inline Quater operator/(const Quater &v, const Quater &w)
{
   return v * w.Inv();
}

inline Quater operator*(const Quater &v, const FLOAT f)
{
   return Quater(v(0)*f, v(1)*f, v(2)*f, v(3)*f);  
}

inline Quater operator*(const FLOAT f, const Quater &v)
{
   return Quater(v(0)*f, v(1)*f, v(2)*f, v(3)*f);  
}

inline Quater operator/(const Quater &v, const FLOAT f)
{
   ASSERT( f != 0.0 );

   return Quater(v(0)/f, v(1)/f, v(2)/f, v(3)/f);  
}

inline Quater operator/(const FLOAT f, const Quater &v)
{
   return f * v.Inv();
}
#endif // __QUATER_H__
